﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;
using System.Windows.Forms;

namespace SimonPG {

    public class TextboxConsole : TextBox {

        OutputStream _Stream;
        TextWriter _Writer;
        TextWriter _OrigWriter;
        TraceListener _TraceListener;
        bool _TrackConsole = false;
        bool _TrackTrace = false;

        public TextboxConsole()
            : base() {

            _Stream = new OutputStream();
            _Stream.OnWritten += _Stream_OnWritten;
            _Writer = new StreamWriterFlush(_Stream);
            _OrigWriter = Console.Out;

            this.TrackConsole = true;
            this.TrackTrace = true;
        }

        [System.ComponentModel.DesignOnly(false)]
        public TextWriter Out { get { return _Writer; } }

        [System.Diagnostics.DebuggerStepThrough]
        void _Stream_OnWritten(byte[] written_data) {

            var text = Encoding.UTF8.GetString(written_data);
            var text_ui = text.Replace("\n", "\r\n");
            this.AppendText(text_ui);

            Console.SetOut(_OrigWriter);
            Console.Write(text);
            Console.SetOut(_Writer);
        }

        public bool TrackConsole {
            get { return _TrackConsole; }
            set {

                if (value) {
                    Console.SetOut(_Writer);
                } else {
                    Console.SetOut(_OrigWriter);
                }
                _TrackConsole = value;
            }
        }

        public bool TrackTrace {
            get { return _TrackTrace; }
            set {

                if (value == _TrackTrace)
                    return;

                if (value) {
                    _TraceListener = new TextWriterTraceListener(this.Out);
                    Trace.Listeners.Add(_TraceListener);
                } else {
                    Trace.Listeners.Remove(_TraceListener);
                    _TraceListener = null;
                }
                _TrackTrace = value;
            }
        }


        [System.Diagnostics.DebuggerStepThrough]
        private class OutputStream : Stream {

            public event Action<byte[]> OnWritten;

            public override void Write(byte[] buffer, int offset, int count) {

                if (this.OnWritten == null)
                    return;

                if (offset == 0 && buffer.Length == count) {
                    this.OnWritten(buffer);

                } else {

                    var buffer_fitted = new byte[count];
                    var list = new List<byte>(buffer);
                    list.CopyTo(offset, buffer_fitted, 0, count);

                    this.OnWritten(buffer_fitted);
                }
            }

            public OutputStream() : base() { }


            // interface for outside

            public override bool CanRead { get { return false; } }
            public override bool CanSeek { get { return false; } }
            public override bool CanWrite { get { return true; } }
            public override void Flush() { }
            public override long Length { get { throw new Exception("This stream has no length."); } }
            public override long Position {
                get { return 0; }
                set { throw new Exception("You can't seek this stream."); }
            }
            public override int Read(byte[] buffer, int offset, int count) { throw new Exception("You can't read from this stream."); }
            public override long Seek(long offset, SeekOrigin origin) { throw new Exception("You can't seek this stream."); }
            public override void SetLength(long value) { throw new Exception("This stream has no length."); }
        }

        [System.Diagnostics.DebuggerStepThrough]
        private class StreamWriterFlush : StreamWriter {

            const int WAIT_TIME = 10; // ms

            Timer _Timer;

            public StreamWriterFlush(Stream stream)
                : base(stream) {
                _Timer = new System.Windows.Forms.Timer();
                // above: The timer should be System.Windows.Forms.Timer,
                // because it runs in the same thread as the parent form, 
                // which is needed to call AppendText in the user control.
                _Timer.Interval = WAIT_TIME;
                _Timer.Tick += _Timer_Tick;
            }

            public override void Write(char value) {
                base.Write(value);
                _Flush();
            }
            public override void Write(char[] buffer) {
                base.Write(buffer);
                _Flush();
            }
            public override void Write(string value) {
                base.Write(value);
                _Flush();
            }
            public override void Write(char[] buffer, int index, int count) {
                base.Write(buffer, index, count);
                _Flush();
            }

            void _Flush() {
                _Timer.Start();
                // above: This section makes performance much higher when
                // writing a few words so many times. It forces to delay 
                // the output to console by using a timer. 
                // Even if you call "Start" many times, the timer occurs 
                // an event "Interval" ms after the first call of "Start" 
                // unless you call "Stop".
            }

            void _Timer_Tick(object sender, EventArgs e) {
                base.Flush();
                _Timer.Stop();
            }
        }
    }
}
