﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;

namespace SimonPG.WaitCore {

    public class EventWaiter {
        // assume: It's only available to wait for EventHandler events.

        readonly IEnumerator<bool> _Etor;
        public EventContainer Container { get; private set; }
        public event EventHandler Finally;
        public bool IsFinished { get; private set; }
        ActionWrapper _OnEndWait = new ActionWrapper();

        public EventWaiter(Func<EventWaiter, IEnumerable<bool>> asyncro_handler) {
            _Etor = asyncro_handler(this).GetEnumerator();
            _MoveEnumeratorNext();
        }

        public bool Wait(EventContainer container) {

            this.Container = container;
            container.Reset();

            var on_end_update = new ActionWrapper();
            container.UpdateEvaluation(on_end_update);
            on_end_update.Call();
            if (container.Evaluation) // if: evaluation is already true before waiting
                return false; // don't wait and go next

            var helper = new WaitHelper();
            helper.EventNotifier = () => {

                on_end_update.Clear();
                container.UpdateEvaluation(on_end_update);
                on_end_update.Call();
                if (container.Evaluation) {

                    _OnEndWait.Call();
                    _MoveEnumeratorNext();
                }
            };
            _OnEndWait.Clear();
            container.SetupWait(helper, _OnEndWait);

            return true;
        }
        public bool Wait(object instance, string event_name) {
            return this.Wait(new Event(instance, event_name));
        }

        void _MoveEnumeratorNext() {

            while (_Etor.MoveNext()) {
                if (_Etor.Current) // if: you wait for event, or don't wait and go next
                    return;
            }

            _Finalize();
        }

        public void CancelWait() {

            _OnEndWait.Call();
            _MoveEnumeratorNext();
        }

        public void CancelFunc() {
            this.CancelWait();
            _Finalize();
        }

        void _Finalize() {

            if (this.Finally != null)
                this.Finally(this, EventArgs.Empty);

            _Etor.Dispose();
            this.IsFinished = true;
        }
    }

    public class ActionWrapper {
        // below: This class is introduced to make it easy to write codes in 
        // the style of "void Start(ActionWrapper on_end) {...}".

        public Action Val;
        public ActionWrapper() { this.Val = null; }
        public void Call() { if (this.Val != null) this.Val(); }
        public void Clear() { this.Val = null; }
    }

    internal class WaitHelper {
        public Action EventNotifier;
        public Dictionary<EventBaseInfo, EventRaiser> EventList =
            new Dictionary<EventBaseInfo, EventRaiser>();
        // above: This Dictionary is introduced to keep only one EventRaiser 
        // created per a Wait call, even when you have many Event instances 
        // which result in waiting the same event.
    }
    internal struct EventBaseInfo {
        public object Instance;
        public string EventName;
        public EventBaseInfo(object instance, string event_name) {
            this.Instance = instance;
            this.EventName = event_name;
        }
    }

    public abstract class EventContainer {

        public virtual bool Evaluation { get; protected set; }

        public virtual void Reset() { }
        internal virtual void SetupWait(WaitHelper helper, ActionWrapper on_end_wait) { }
        internal abstract bool UpdateEvaluation(ActionWrapper on_end_update);
    }

    public class Event : EventContainer {
        // assume: It's only available to wait for EventHandler events.

        public readonly object Instance;
        public readonly EventInfo EventInfo;
        EventRaiser _Raiser = EventRaiser.Empty;
        public object Sender { get { return _Raiser.Sender; } }
        public EventArgs E { get { return _Raiser.E; } }
        public override bool Evaluation { get { return _Raiser.Evaluation; } }

        public Event(object instance, string event_name) {
            this.Instance = instance;
            this.EventInfo = instance.GetType().GetEvent(event_name);
        }

        public override void Reset() { _Raiser.Reset(); }

        internal override void SetupWait(WaitHelper helper, ActionWrapper on_end_wait) {

            var event_info = new EventBaseInfo() {
                Instance = this.Instance, EventName = this.EventInfo.Name
            };
            if (helper.EventList.ContainsKey(event_info)) { // if: the same event is already setup
                _Raiser = helper.EventList[event_info];

            } else { // if: it's a new type of event
                _Raiser = new EventRaiser(this);
                _Raiser.SetupEvent(helper, on_end_wait);
                helper.EventList.Add(event_info, _Raiser);
            }
        }

        internal override bool UpdateEvaluation(ActionWrapper on_end_update) { return _Raiser.Evaluation; }
    }

    internal class EventRaiser {
        // assume: It's only available to wait for EventHandler events.

        Event _Event;
        public object Sender { get; private set; }
        public EventArgs E { get; private set; }
        public bool Evaluation { get; private set; }

        public EventRaiser(Event event_) { _Event = event_; }

        public virtual void Reset() {
            this.Sender = null;
            this.E = null;
            this.Evaluation = false;
        }

        public virtual void SetupEvent(WaitHelper helper, ActionWrapper on_end_wait) {

            EventHandler handler = null;
            var is_removed = false;
            handler = (_sender, _e) => {

                if (is_removed)
                    return;
                // above: When you remove event handler using GetRemoveMethod, it doesn't 
                // remove it immediately, so you have to check if it has been removed or 
                // not here.

                if (this.Evaluation)
                    return;

                this.Sender = _sender;
                this.E = _e;
                this.Evaluation = true;
                helper.EventNotifier();
            };
            _Event.EventInfo.GetAddMethod().Invoke(_Event.Instance, new[] { handler });

            on_end_wait.Val += () => {

                _Event.EventInfo.GetRemoveMethod().Invoke(_Event.Instance, new[] { handler });
                is_removed = true;
            };
        }

        static EventRaiserEmpty _Empty = new EventRaiserEmpty();
        public static EventRaiser Empty { get { return _Empty; } }
    }
    internal class EventRaiserEmpty : EventRaiser {
        public EventRaiserEmpty() : base(null) { }
        public override void Reset() { }
        public override void SetupEvent(WaitHelper helper, ActionWrapper on_end_wait) { }
    }

    public abstract class Condition : EventContainer {

        protected readonly EventContainer[] _Containers;

        protected Condition(params EventContainer[] containers) { _Containers = containers; }

        public sealed override void Reset() {
            if (_Containers != null)
                _Containers.ForEach(c => c.Reset());
            this.Evaluation = false;
            this.OnReset();
        }

        internal sealed override void SetupWait(WaitHelper helper, ActionWrapper on_end_wait) {
            if (_Containers != null) {
                _Containers.ForEach(c => c.SetupWait(helper, on_end_wait));
            }
            this.OnSetupWait(on_end_wait);
        }

        internal sealed override bool UpdateEvaluation(ActionWrapper on_end_update) {
            this.Evaluation = this.OnUpdateEvaluation(on_end_update);
            return this.Evaluation;
        }

        protected virtual void OnReset() { }
        protected virtual void OnSetupWait(ActionWrapper on_end_wait) { }
        protected abstract bool OnUpdateEvaluation(ActionWrapper on_end_update);
    }

    public class And : Condition {

        public EventContainer[] Containers { get { return _Containers; } }

        public And(params EventContainer[] containers) : base(containers) { }

        protected override bool OnUpdateEvaluation(ActionWrapper on_end_update) {
            var is_raised = true;
            this.Containers.ForEach(c => is_raised &= c.UpdateEvaluation(on_end_update));
            return is_raised;
        }
    }

    public class Or : Condition {

        public EventContainer[] Containers { get { return _Containers; } }

        public Or(params EventContainer[] containers) : base(containers) { }

        protected override bool OnUpdateEvaluation(ActionWrapper on_end_update) {
            var is_raised = false;
            this.Containers.ForEach(c => is_raised |= c.UpdateEvaluation(on_end_update));
            return is_raised;
        }
    }

    public class Not : Condition {

        public EventContainer Container { get { return _Containers.First(); } }

        public Not(EventContainer container) : base(container) { }

        protected override bool OnUpdateEvaluation(ActionWrapper on_end_update) {
            return !this.Container.UpdateEvaluation(on_end_update);
        }
    }

    public class Count : Condition {

        public int CurrentCount;
        public int WaitCount;

        public EventContainer Container { get { return _Containers.First(); } }

        public Count(EventContainer container, int wait_count) : base(container) { this.WaitCount = wait_count; }

        protected override void OnReset() { this.CurrentCount = 0; }

        protected override bool OnUpdateEvaluation(ActionWrapper on_end_update) {

            if (this.CurrentCount >= this.WaitCount)
                return true;

            if (this.Container.Evaluation) {

                on_end_update.Val += this.Container.Reset;
                // above: Call Reset() after all evaluation update has been done.
                // Otherwise, the wait connection of (Count(e1, 3) and Count(e1, 5)) 
                // cannot wait for e1 raising 5 times.

                this.CurrentCount++;
                if (this.CurrentCount >= this.WaitCount)
                    return true;
            }
            return false;
        }
    }

    public class Eval : Condition {

        public EventContainer Container { get { return _Containers.First(); } }
        public Func<bool> Evaluator { get; set; }

        public Eval(EventContainer container, Func<bool> evaluator)
            : base(container) { this.Evaluator = evaluator; }

        protected override bool OnUpdateEvaluation(ActionWrapper on_end_update) {

            var eval_event = this.Container.Evaluation;
            var eval_logic = this.Evaluator();
            if (eval_event && !eval_logic)
                this.Container.Reset();
            // above: You need to reset all in order to raise event again to 
            // re-evaluate Eval when the output of Eval is false.

            return eval_event && eval_logic;
        }
    }


    // Test code

    internal class Test_EventWaiting {

        System.Windows.Forms.Timer _Timer;

        public void Test_WaitUsingYield() {

            Console.WriteLine("\n------------------------------");
            Console.WriteLine("Started \"{0}\"", System.Reflection.MethodBase.GetCurrentMethod().Name);

            _Timer = new System.Windows.Forms.Timer() {
                Interval = 500,
                Enabled = true,
            };
            _Timer.Tick += new EventHandler(_Timer_Tick);

            new EventWaiter(_WaitUsingYieldHandler);
        }

        void _Timer_Tick(object sender, EventArgs e) {
            Console.WriteLine("Tick of Timer is raised");
        }

        IEnumerable<bool> _WaitUsingYieldHandler(EventWaiter waiter) {

            Console.WriteLine("Test Event class and wait.");
            for (var i = 0; i <= 2; i++) {

                Console.WriteLine("{0}", i);

                yield return waiter.Wait(new Event(_Timer, "Tick"));
            }

            Console.WriteLine("Test Event class and wait with the same instance.");
            var timer_tick = new Event(_Timer, "Tick");
            for (var i = 0; i <= 2; i++) {

                Console.WriteLine("{0}", i);

                yield return waiter.Wait(_Timer, "Tick");
            }

            Console.WriteLine("Finished single cycle.");
            _Timer.Dispose();
        }
    }

    internal class Test_AndOrWaiting {

        System.Windows.Forms.Timer _TimerL;
        System.Windows.Forms.Timer _TimerM;
        System.Windows.Forms.Timer _TimerH;

        public void Test_AndOrConnection() {

            Console.WriteLine("\n------------------------------");
            Console.WriteLine("Started \"{0}\"", System.Reflection.MethodBase.GetCurrentMethod().Name);

            _TimerL = new System.Windows.Forms.Timer() {
                Interval = 1100,
                Enabled = true,
                Tag = "L",
            };
            _TimerM = new System.Windows.Forms.Timer() {
                Interval = 550,
                Enabled = true,
                Tag = "M",
            };
            _TimerH = new System.Windows.Forms.Timer() {
                Interval = 220,
                Enabled = true,
                Tag = "H",
            };
            _TimerL.Tick += new EventHandler(_Timers_Tick);
            _TimerM.Tick += new EventHandler(_Timers_Tick);
            _TimerH.Tick += new EventHandler(_Timers_Tick);

            new EventWaiter(_AndOrHandler);
        }

        void _Timers_Tick(object sender, EventArgs e) {
            Console.WriteLine("Tick of Timer({0}) is raised", ((System.Windows.Forms.Timer)sender).Tag);
        }

        IEnumerable<bool> _AndOrHandler(EventWaiter waiter) {

            var timer_tick_L = new Event(_TimerL, "Tick");
            var timer_tick_M = new Event(_TimerM, "Tick");
            var timer_tick_H = new Event(_TimerH, "Tick");

            Console.WriteLine("Test (L and M and H) connection.");
            for (var i = 0; i <= 2; i++) {

                Console.WriteLine("{0}", i);

                yield return waiter.Wait(new And(timer_tick_L, timer_tick_M, timer_tick_H));
            }

            Console.WriteLine("Test (L or M or H) connection.");
            for (var i = 0; i <= 2; i++) {

                Console.WriteLine("{0}", i);

                yield return waiter.Wait(new Or(timer_tick_L, timer_tick_M, timer_tick_H));
            }

            Console.WriteLine("Test (L and M and H) connection with the same instance.");
            var and_timer_tick_LMH = new And(timer_tick_L, timer_tick_M, timer_tick_H);
            for (var i = 0; i <= 2; i++) {

                Console.WriteLine("{0}", i);

                yield return waiter.Wait(and_timer_tick_LMH);
            }

            Console.WriteLine("Test (L or M or H) connection with the same instance.");
            var or_timer_tick_LMH = new Or(timer_tick_L, timer_tick_M, timer_tick_H);
            for (var i = 0; i <= 2; i++) {

                Console.WriteLine("{0}", i);

                yield return waiter.Wait(or_timer_tick_LMH);
            }

            Console.WriteLine("Finished single cycle.");
            _TimerL.Dispose();
            _TimerM.Dispose();
            _TimerH.Dispose();
        }

        public void Test_AdvancedConnection() {

            Console.WriteLine("\n------------------------------");
            Console.WriteLine("Started \"{0}\"", System.Reflection.MethodBase.GetCurrentMethod().Name);

            _TimerL = new System.Windows.Forms.Timer() {
                Interval = 660,
                Enabled = true,
                Tag = "L",
            };
            _TimerM = new System.Windows.Forms.Timer() {
                Interval = 550,
                Enabled = true,
                Tag = "M",
            };
            _TimerH = new System.Windows.Forms.Timer() {
                Interval = 440,
                Enabled = true,
                Tag = "H",
            };
            _TimerL.Tick += new EventHandler(_Timers_Tick);
            _TimerM.Tick += new EventHandler(_Timers_Tick);
            _TimerH.Tick += new EventHandler(_Timers_Tick);

            new EventWaiter(_AdvancedHandler);
        }

        IEnumerable<bool> _AdvancedHandler(EventWaiter waiter) {

            var timer_tick_L = new Event(_TimerL, "Tick");
            var timer_tick_M = new Event(_TimerM, "Tick");
            var timer_tick_H = new Event(_TimerH, "Tick");

            Console.WriteLine("Test advanced connection of (L and (M or H)).");
            var timer_tick_LMH1 = new And(timer_tick_L, new Or(timer_tick_M, timer_tick_H));
            for (var i = 0; i <= 5; i++) {

                Console.WriteLine("{0}", i);

                yield return waiter.Wait(timer_tick_LMH1);
            }

            Console.WriteLine("Test advanced connection of ((L or M) and H).");
            var timer_tick_LMH2 = new And(new Or(timer_tick_L, timer_tick_M), timer_tick_H);
            for (var i = 0; i <= 5; i++) {

                Console.WriteLine("{0}", i);

                yield return waiter.Wait(timer_tick_LMH2);
            }

            Console.WriteLine("Finished single cycle.");
            _TimerL.Dispose();
            _TimerM.Dispose();
            _TimerH.Dispose();
        }
    }

    internal class Test_NotWaiting {

        System.Windows.Forms.Timer _TimerL;
        System.Windows.Forms.Timer _TimerM;
        System.Windows.Forms.Timer _TimerH;

        public void Test_NotConnection() {

            Console.WriteLine("\n------------------------------");
            Console.WriteLine("Started \"{0}\"", System.Reflection.MethodBase.GetCurrentMethod().Name);

            _TimerM = new System.Windows.Forms.Timer() {
                Interval = 550,
                Enabled = true,
                Tag = "M",
            };
            _TimerM.Tick += new EventHandler(_Timers_Tick);

            new EventWaiter(_NotHandler);
        }

        void _Timers_Tick(object sender, EventArgs e) {
            Console.WriteLine("Tick of Timer({0}) is raised", ((System.Windows.Forms.Timer)sender).Tag);
        }

        IEnumerable<bool> _NotHandler(EventWaiter waiter) {

            var timer_tick_M = new Event(_TimerM, "Tick");

            Console.WriteLine("Test (not M) connection.");
            yield return waiter.Wait(new Not(timer_tick_M));

            Console.WriteLine("Test (not (not M)) connection.");
            yield return waiter.Wait(new Not(new Not(timer_tick_M)));

            Console.WriteLine("Test (not (not (not M))) connection.");
            yield return waiter.Wait(new Not(new Not(new Not(timer_tick_M))));

            Console.WriteLine("Test (not (not (not (not M)))) connection.");
            yield return waiter.Wait(new Not(new Not(new Not(new Not(timer_tick_M)))));

            Console.WriteLine("Finished single cycle.");
            _TimerM.Dispose();
        }

        public void Test_AdvancedConnection() {

            Console.WriteLine("\n------------------------------");
            Console.WriteLine("Started \"{0}\"", System.Reflection.MethodBase.GetCurrentMethod().Name);

            _TimerL = new System.Windows.Forms.Timer() {
                Interval = 660,
                Enabled = true,
                Tag = "L",
            };
            _TimerM = new System.Windows.Forms.Timer() {
                Interval = 550,
                Enabled = true,
                Tag = "M",
            };
            _TimerH = new System.Windows.Forms.Timer() {
                Interval = 440,
                Enabled = true,
                Tag = "H",
            };
            _TimerL.Tick += new EventHandler(_Timers_Tick);
            _TimerM.Tick += new EventHandler(_Timers_Tick);
            _TimerH.Tick += new EventHandler(_Timers_Tick);

            new EventWaiter(_AdvancedHandler);
        }

        IEnumerable<bool> _AdvancedHandler(EventWaiter waiter) {

            var timer_tick_L = new Event(_TimerL, "Tick");
            var timer_tick_M = new Event(_TimerM, "Tick");
            var timer_tick_H = new Event(_TimerH, "Tick");

            Console.WriteLine("Test advanced connection of ((H and not M) or L).");
            var timer_tick_LMH1 = new Or(new And(timer_tick_H, new Not(timer_tick_M)), timer_tick_L);
            for (var i = 0; i <= 5; i++) {

                Console.WriteLine("{0}", i);

                yield return waiter.Wait(timer_tick_LMH1);
            }

            Console.WriteLine("Test advanced connection of (not (not H or not M or not L)).");
            var timer_tick_LMH2 = new Not(new Or(new Not(timer_tick_H), new Not(timer_tick_M), new Not(timer_tick_L)));
            for (var i = 0; i <= 5; i++) {

                Console.WriteLine("{0}", i);

                yield return waiter.Wait(timer_tick_LMH2);
            }

            Console.WriteLine("Finished single cycle.");
            _TimerL.Dispose();
            _TimerM.Dispose();
            _TimerH.Dispose();
        }
    }

    internal class Test_Canceling {

        System.Windows.Forms.Timer _TimerL;
        System.Windows.Forms.Timer _TimerH;

        public void Test_CancelWait() {

            Console.WriteLine("\n------------------------------");
            Console.WriteLine("Started \"{0}\"", System.Reflection.MethodBase.GetCurrentMethod().Name);

            _TimerL = new System.Windows.Forms.Timer() {
                Interval = 2000,
                Enabled = true,
                Tag = "L",
            };
            _TimerH = new System.Windows.Forms.Timer() {
                Interval = 500,
                Enabled = true,
                Tag = "H",
            };
            _TimerL.Tick += new EventHandler(_Timers_Tick);
            _TimerH.Tick += new EventHandler(_Timers_Tick);

            var waiter = new EventWaiter(_CancelWaitHandler);

            _TimerL.Tick += (_s, _e) => {

                Console.WriteLine("Cancel waiting.");
                waiter.CancelWait();

                if (waiter.IsFinished)
                    _TimerL.Dispose();
            };
        }

        void _Timers_Tick(object sender, EventArgs e) {
            Console.WriteLine("Tick of Timer({0}) is raised", ((System.Windows.Forms.Timer)sender).Tag);
        }

        IEnumerable<bool> _CancelWaitHandler(EventWaiter waiter) {

            var timer_tick_H = new Event(_TimerH, "Tick");
            waiter.Finally += (_s, _e) => { _TimerH.Dispose(); };

            Console.WriteLine("Test (H and not H) connection. (infinite waiting)");
            yield return waiter.Wait(new And(timer_tick_H, new Not(timer_tick_H)));

            Console.WriteLine("Test (not H and H) connection. (infinite waiting)");
            yield return waiter.Wait(new And(new Not(timer_tick_H), timer_tick_H));

            Console.WriteLine("Test (H and not H) connection creating Event each time. (infinite waiting)");
            yield return waiter.Wait(new And(new Event(_TimerH, "Tick"), new Not(new Event(_TimerH, "Tick"))));

            Console.WriteLine("Test (not H and H) connection creating Event each time. (infinite waiting)");
            yield return waiter.Wait(new And(new Not(new Event(_TimerH, "Tick")), new Event(_TimerH, "Tick")));

            Console.WriteLine("Finished single cycle.");
        }
        public void Test_CancelFunc() {

            Console.WriteLine("\n------------------------------");
            Console.WriteLine("Started \"{0}\"", System.Reflection.MethodBase.GetCurrentMethod().Name);

            _TimerL = new System.Windows.Forms.Timer() {
                Interval = 2000,
                Enabled = true,
                Tag = "L",
            };
            _TimerH = new System.Windows.Forms.Timer() {
                Interval = 500,
                Enabled = true,
                Tag = "H",
            };
            _TimerL.Tick += new EventHandler(_Timers_Tick);
            _TimerH.Tick += new EventHandler(_Timers_Tick);

            var waiter = new EventWaiter(_CancelFuncHandler);

            _TimerL.Tick += (_s, _e) => {

                Console.WriteLine("Cancel enumerator action.");
                waiter.CancelFunc();

                _TimerL.Dispose();
            };
        }

        IEnumerable<bool> _CancelFuncHandler(EventWaiter waiter) {

            var timer_tick_H = new Event(_TimerH, "Tick");
            waiter.Finally += (_s, _e) => { _TimerH.Dispose(); };

            while (true) {
                Console.WriteLine("Test (H) waiting.");
                yield return waiter.Wait(timer_tick_H);
            }

            Console.WriteLine("Finished single cycle.");
        }
    }

    internal class Test_UserWaiting {

        System.Windows.Forms.Timer _Timer;

        public void Test_CountWaiting() {

            Console.WriteLine("\n------------------------------");
            Console.WriteLine("Started \"{0}\"", System.Reflection.MethodBase.GetCurrentMethod().Name);

            _Timer = new System.Windows.Forms.Timer() {
                Interval = 500,
                Enabled = true,
            };
            _Timer.Tick += new EventHandler(_Timer_Tick);

            new EventWaiter(_CountWaitingHandler);
        }

        void _Timer_Tick(object sender, EventArgs e) {
            Console.WriteLine("Tick of Timer is raised");
        }

        IEnumerable<bool> _CountWaitingHandler(EventWaiter waiter) {

            var timer_tick = new Event(_Timer, "Tick");
            Console.WriteLine("Test (Tick[3]) connection. ([x] = wait x times)");
            yield return waiter.Wait(new Count(timer_tick, 3));

            Console.WriteLine("Test (Tick and Tick), then (Tick[3]) connection. ([x] = wait x times)");
            yield return waiter.Wait(new And(timer_tick, timer_tick));
            yield return waiter.Wait(new Count(timer_tick, 3));

            Console.WriteLine("Test (Tick[3] and Tick[5]) connection creating Event each time. ([x] = wait x times)");
            yield return waiter.Wait(new And(
                new Count(new Event(_Timer, "Tick"), 3), new Count(new Event(_Timer, "Tick"), 5)));

            Console.WriteLine("Test (Tick[3] and Tick[5]) connection with the same instance. ([x] = wait x times)");
            yield return waiter.Wait(new And(new Count(timer_tick, 3), new Count(timer_tick, 5)));

            Console.WriteLine("Finished single cycle.");
            _Timer.Dispose();
        }
    }

    internal class Test_EvalWaiting {

        System.Windows.Forms.Timer _Timer;

        public void Test_WaitWithEval() {

            Console.WriteLine("\n------------------------------");
            Console.WriteLine("Started \"{0}\"", System.Reflection.MethodBase.GetCurrentMethod().Name);

            _Timer = new System.Windows.Forms.Timer() {
                Interval = 300,
                Enabled = true,
            };
            _Timer.Tick += new EventHandler(_Timer_Tick);

            new EventWaiter(_WaitWithEvalHandler);
        }

        void _Timer_Tick(object sender, EventArgs e) {
            Console.WriteLine("Tick of Timer is raised");
        }

        IEnumerable<bool> _WaitWithEvalHandler(EventWaiter waiter) {

            Console.WriteLine("Test Eval(Tick and Random >= 5).");
            var rand = new Random();
            var rand_val = 0;
            Func<int> get_rand_10 = () => {
                rand_val = rand.Next(10);
                return rand_val;
            };
            var tick_and_even = new Eval(new Event(_Timer, "Tick"), () => get_rand_10() >= 5);
            for (var i = 0; i <= 3; i++) {

                yield return waiter.Wait(tick_and_even);
                Console.WriteLine("Random: {0}. (should be >= 5)", rand_val);
            }

            Console.WriteLine("Finished single cycle.");
            _Timer.Dispose();
        }
    }


    // Additional extention method

    internal static class ExtentionMethods {

        [System.Diagnostics.DebuggerStepThrough]
        public static void ForEach<T>(this IEnumerable<T> enumerator, Action<T> action) {
            foreach (var item in enumerator) {
                action(item);
            }
        }
        [System.Diagnostics.DebuggerStepThrough]
        public static void ForEach<T>(this IEnumerable<T> enumerator, Action<T, int> action) {
            var i = 0;
            foreach (var item in enumerator) {
                action(item, i);
                i++;
            }
        }

        [System.Diagnostics.DebuggerStepThrough]
        public static T First<T>(this IList<T> list) { return list[0]; }
        [System.Diagnostics.DebuggerStepThrough]
        public static T Last<T>(this IList<T> list) { return list[list.Count - 1]; }
    }
}
